from Resource import *

class AbstractModel(object):
    """
    Base class defining the operation of the Mugshot data model.

    AbstractModel defines to core operation of the Mugshot data model.
    Normal usage is to use the singleton instance of the DataModel
    subclass, which connects the DataModel to the user's Mugshot
    session.  Other subclasses are used for testing purposes.
    
    """

    def __init__(self):
        self.__initialized_handlers = []
        self.__connected_handlers = []
        self.__server_connected_handlers = []
        self.__disconnected_handlers = []
        self.__added_handlers = []
        self.__removed_handlers = []
        self.__resources = {}
        self.initialized = False
        # self_id that is not None means that the model is connected
        self.self_id = None
        # connected is True only when we can talk to the server/ the network is present
        self.connected = False
        self.__handled_model_connected = False
        self.__handled_server_connected = False

    def add_initialized_handler(self, handler):
        """Add a handler that will be called when we initialize the model and find out if we can get connected to the server"""
        self.__initialized_handlers.append(handler)

    def remove_initialized_handler(self, handler):
        """Remove a handler added with add_initialized_handler"""
        self.__initialized_handlers.remove(handler)

    def add_connected_handler(self, handler):
        """Add a handler that will be called when we become connected to the model"""
        self.__connected_handlers.append(handler)

    def remove_connected_handler(self, handler):
        """Remove a handler added with add_connected_handler"""
        self.__connected_handlers.remove(handler)

    def add_server_connected_handler(self, handler):
        """Add a handler that will be called when we become connected to the server"""
        self.__server_connected_handlers.append(handler)

    def remove_server_connected_handler(self, handler):
        """Remove a handler added with add_connected_handler"""
        self.__server_connected_handlers.remove(handler)

    def add_disconnected_handler(self, handler):
        """Add a handler that will be called when we become disconnected from the server"""
        self.__disconnected_handlers.append(handler)

    def remove_disconnected_handler(self, handler):
        """Remove a handler added with add_disconnected_handler"""
        self.__disconnected_handlers.remove(handler)

    def add_added_handler(self, handler):
        """Add a handler that will be called when a resource is added"""
        self.__added_handlers.append(handler)

    def remove_added_handler(self, handler):
        """Remove a handler added with add_added_handler"""
        self.__added_handlers.remove(handler)

    ## of course, currently we never remove resources...
    def add_removed_handler(self, handler):
        """Add a handler that will be called when a resource is removed"""
        self.__removed_handlers.append(handler)

    def remove_removed_handler(self, handler):
        """Remove a handler added with add_removed_handler"""
        self.__removed_handlers.remove(handler)

    def get_resources(self):
        """Get currently-known resources; keep in mind, it's totally undefined what we might currently know, it just depends on what queries people have done"""
        return self.__resources.values()

    def query(self, method, fetch=None, single_result=False, **kwargs):
        """Create a query object.

        Arguments:
        method -- the method to call. A tuple of (namespace_uri, name)
        fetch -- the fetch string to use for retrieving data. (default '+')
        single_result -- if True, the method is expected to return exactly one object, and an
           error will result if it returns 0 or more than 1 object. If False,
           the method returns a possibly-empty list of objects. (default False)

        Additional keyword arguments are passed to the query
        method. The query will not actually be sent until you call
        execute() on it.
        
        """
        raise NotImplementedException()

    def query_resource(self, resource_id, fetch):
        """Create a query object for the standard system method 'getResource'.

        Arguments:
        method -- the ID of the resource to retrieve
        fetch -- the fetch string to use for retrieving data. (default '+')

        The query will return the resource object if found, otherwise
        an error will result. The query will not actually be sent
        until you call execute() on it.
        
        """
        
        return self.query(("http://mugshot.org/p/system", "getResource"),
                          fetch,
                          single_result=True,
                          resourceId=resource_id)
                       
    def update(self, method,  **kwargs):
        """Create a query object for an update method

        Arguments:
        method -- the method to call. A tuple of (namespace_uri, name)

        Additional keyword arguments are passed to the query
        method. The query will not actually be sent until you call
        execute() on it. Update methods don't have a result, but you can
        still add handlers to be called on success or error.
        
        """
        
        raise NotImplementedException()

    def _get_resource(self, resource_id):
        return self.__resources[resource_id]

    def _ensure_resource(self, resource_id, class_id):
        try:
            return self.__resources[resource_id]
        except KeyError:
            resource = self.__resources[resource_id] = Resource(self, resource_id, class_id)

            for handler in self.__added_handlers:
                handler(resource)
            
            return resource

    def _on_initialized(self):
        for handler in self.__initialized_handlers:
            handler()    
        self.initialized = True

    def _on_connected(self):
        if not self.__handled_model_connected:
      
            # On reconnection, all previous state is irrelevant
            self.__resources = {}
        
            for handler in self.__connected_handlers:
                handler()

            self.__handled_model_connected = True 

        if self.connected:
            self._on_server_connected()

    def _on_server_connected(self):
        if self.__handled_server_connected:
            return

        for handler in self.__server_connected_handlers:
            handler()

        self.__handled_server_connected = True

    def _on_disconnected(self):
        if not self.__handled_model_connected:
            return
       
        for handler in self.__disconnected_handlers:
            handler()

        self.__handled_model_connected = False
        self.__handled_server_connected = False

    def _set_self_id(self, self_id):
        self.self_id = self_id
